// Heart trails.  Uses ideas from Thomas Jakobsen's paper "Advanced
// Character Physics" see http://www.ioi.dk/~ti

#include "main.h"
#include "heart_trails.h"
#include "level.h"
#include "heart.h"
#include "rocket.h"
#include "arrow.h"

bool heart_trails = true;

const float gTimeStep = (1.0 / 70.0);

template<int N>
class PhysParticles
{
public:
    Vector mPos[N];
    Vector mOldPos[N];
    Vector mAccel[N];

    PhysParticles()
    {
    }

    void update()
    {
        reset_forces();
        accumulate_forces();
        integrate();
        satisfy_constraints();
    }


    void reset_forces()
    {
        for(int i=0; i<N; i++) {
            mAccel[i] = Vector(0,0);
        }
    }

    virtual void accumulate_forces()
    {
    }


    void distance_constraint(Vector& p1, Vector& p2, float dist, float fixamount=1)
    {
        Vector d = p2 - p1;
        float dl = d.magnitude();
        
        if(dl == 0)
            return;

        float diff = (dl - dist) / dl;
        
        p1 += d * (0.5 * fixamount) * diff;
        p2 -= d * (0.5 * fixamount) * diff;
    }


    void cond_distance_constraint(Vector& p1, Vector& p2, float mindist) 
    {
        Vector d = p2 - p1;
        float dl = d.magnitude();

        if(dl < mindist) {
            float diff = (dl - mindist) / dl;
            
            p1 += d * 0.5 * diff;
            p2 -= d * 0.5 * diff;
        }
    }

    virtual void satisfy_constraints()
    {
    }

    void integrate()
    {
        for(int i=0; i<N; i++) {
            Vector& x = mPos[i];
            Vector temp = x;
            Vector& oldx = mOldPos[i];
            Vector& a = mAccel[i];

            x += x - oldx + a * gTimeStep * gTimeStep;

            oldx = temp;
        }
    }
};


const int NUM_PARTICLES = 10;

class HeartTrails : public PhysParticles<NUM_PARTICLES>
{
public:
    HeartTrails()
    {
        mSnapped = false;
    }


    virtual void accumulate_forces()
    {
        // some force for now
        for(int i=0; i<NUM_PARTICLES; i++) {
            mAccel[i] += Vector(0, 200) - (mPos[i] - mOldPos[i]) * 100;
        }
    }


    template<class T>
    void thing_collision_constraint(ThingList<T>& things, Vector& TL, Vector& BR, float radmul=1.0, bool sniptrail=false)
    {
        // collide against things - which are conveniently sorted by their x coord
        ThingList<T>::iterator it;

        it = things.begin();

        // skip all things to the left of the bounding box
        while(it != things.end()) {
            T& thing = *it;

            Vector& p = thing.mPos;
            float r = thing.mRadius * radmul;

            if(p.x + r >= TL.x)
                break;

            it++;
        }

        // now check all other rockets, until we find the first
        // one to the right of the bounding box

        while(it != things.end()) {
            T& thing = *it;

            Vector& rp = thing.mPos;
            float r = thing.mRadius * radmul;

            if(rp.x + r >= TL.x &&
               rp.x - r <= BR.x &&
               rp.y + r >= TL.y &&
               rp.y - r <= BR.y) {

                // this rocket is inside the bounding box
                // so test against all particles in the string
                for(int i=0; i<NUM_PARTICLES; i++) {
                    Vector& p = mPos[i];

                    if(rp.x + r < p.x) continue;
                    if(rp.x - r > p.x) continue;
                    if(rp.y + r < p.y) continue;
                    if(rp.y - r > p.y) continue;

                    Vector d = rp - p;
                        
                    if(d.square_magnitude() < r*r) {
                        p = p - (~d) * r * 0.5;
                        if(sniptrail && !mSnapped) {
                            mSnapped = true;
                            mSnapTimer = 70 * 3;
                            mSnapIndex = MID(1,i+1, NUM_PARTICLES-3);
                        }
                    }
                }

            }
            else if(rp.x - r > BR.y) {
                // this rocket's left side is to the right of the bounding box
                break;
            }

            it++;
        }
            

    }

    virtual void satisfy_constraints()
    {
        int i;

        for(int j=0; j<3; j++) {

            if(!mSnapped) {
                // keep string as ... string
                for(i=0; i<NUM_PARTICLES-1; i++)
                    distance_constraint(mPos[i], mPos[i+1], 6, 1);
            }
            else {
                // snapped string
                mSnapIndex = MID(1, mSnapIndex, NUM_PARTICLES-1);

                for(i=0; i<mSnapIndex-1; i++)
                    distance_constraint(mPos[i], mPos[i+1], 10, 1);

                for(i=mSnapIndex; i<NUM_PARTICLES-1; i++)
                    distance_constraint(mPos[i], mPos[i+1], 10, 1);
            }


            // work out a bounding box
            Vector TL, BR;

            TL = BR = mPos[0];
            for(i=1; i<NUM_PARTICLES; i++) {
                Vector& p = mPos[i];

                TL.x = MIN(TL.x, p.x);
                TL.y = MIN(TL.y, p.y);

                BR.x = MAX(BR.x, p.x);
                BR.y = MAX(BR.y, p.y);
            }


            thing_collision_constraint(rockets, TL, BR);
            thing_collision_constraint(arrows, TL, BR, 5, true);

            keep_points_inside_world(mOldPos, mPos, NUM_PARTICLES);

            // stop string going through the center for the heart
            for(int i=0; i<NUM_PARTICLES; i++) {
                Vector c = heart->mPos;
                cond_distance_constraint(c, mPos[i], heart->mRadius);
            }
            
            mTugPos = mPos[0];
            mPos[0] = mAttachPos;
        }

        // check for snaps
        if(!mSnapped) {
            for(i=0; i<NUM_PARTICLES-1; i++) {
                Vector d = mPos[i+1] - mPos[i];

                if(d.square_magnitude() > 2000) {
                    mSnapped = true;
                    mSnapTimer = 70 * 2;
                    mSnapIndex = MID(1, i+1, NUM_PARTICLES-3);
                    break;
                }
            }
        }
        else {
            mSnapTimer--;
            if(mSnapTimer <= 0) {
                mSnapped=false;
                for(i=mSnapIndex; i<NUM_PARTICLES; i++) {
                    mOldPos[i] = mPos[i] = mPos[mSnapIndex-1];
                }
            }
        }
    }

    Vector mAttachPos;
    Vector mTugPos;

    bool mSnapped;
    int mSnapIndex;
    int mSnapTimer;
};


static HeartTrails trails[2];


void init_heart_trails()
{
    Vector right(-heart->mFront.y, heart->mFront.x);

    trails[0].mAttachPos = heart->mPos - (heart->mFront * 20) + (right * 20);
    trails[1].mAttachPos = heart->mPos - (heart->mFront * 20) - (right * 20);

    for(int i=0; i<2; i++) {
        for(int j=0; j<NUM_PARTICLES; j++) {
            trails[i].mPos[j] = trails[i].mOldPos[j] = trails[i].mAttachPos;
        }
    }
}


bool update_heart_trails()
{
    if(!heart_trails)
        return true;

    Vector right(-heart->mFront.y, heart->mFront.x);

    trails[0].mAttachPos = heart->mPos - (heart->mFront * 20) + (right * 10);
    trails[1].mAttachPos = heart->mPos - (heart->mFront * 20) - (right * 10);
        

    trails[0].update();
    trails[1].update();

    heart->mVel += (trails[0].mTugPos - trails[0].mAttachPos) * 0.002 + (trails[1].mTugPos - trails[1].mAttachPos) * 0.002;

    return true;
}


void render_heart_trails(BITMAP* bmp)
{
    if(!heart_trails)
        return;

    for(int i=0; i<2; i++) {
        HeartTrails& trail = trails[i];

        if(!trail.mSnapped) {
            for(int j=0; j<NUM_PARTICLES-1; j++) {
                Vector p1 = trail.mPos[j] + scroll;
                Vector p2 = trail.mPos[j+1] + scroll;
                
                line(bmp, p1.x, p1.y, p2.x, p2.y, makecol(255, 128,0));
                
            }
        }
        else {
            for(int j=0; j<trail.mSnapIndex-1; j++) {
                Vector p1 = trail.mPos[j] + scroll;
                Vector p2 = trail.mPos[j+1] + scroll;
                
                line(bmp, p1.x, p1.y, p2.x, p2.y, makecol(255, 128,0));
            }

            for(int j=trail.mSnapIndex; j<NUM_PARTICLES-1; j++) {
                Vector p1 = trail.mPos[j] + scroll;
                Vector p2 = trail.mPos[j+1] + scroll;
                
                line(bmp, p1.x, p1.y, p2.x, p2.y, makecol(255, 128,0));
            }
                                                                 
        }
    }
}
