// the particle objects

#include "particle.h"
#include "sound.h"
#include "level.h"
#include "palette.h"
#include "flower.h"



// enable/disable particles
int use_particles = 2;


// global list of all particles
ThingList<Particle> particles;


// flowers
Flower *flower_red_yellow, *flower_blue_orange, *flower_purple_red, *flower_blue_purple, *flower_magenta_violet;


// colours
int yellow, white;



// create a particle
Particle::Particle (const Vector &pos, const Vector &vel, int lifespan, float gravity, int type, bool collide, Flower *flower)
{
	mOldPos = mPos = pos;
	mVel = vel;
	mLifeSpan = lifespan;
	mGravity = gravity;
	mType = type;
	mFlower = flower;
	mCollide = collide && (use_particles >= 2);
}



// particle update function
bool Particle::update()
{
	mOldPos = mPos;
	mPos += mVel;
	mVel.y += mGravity;

	if (mCollide) {
		vector<LevelCollision> collisions;
	
		if (check_level_collision_line (mOldPos, mPos, collisions)) {
			mPos += collisions[0].normal * collisions[0].dist;
			mVel -= collisions[0].normal * 1.5 * (mVel | collisions[0].normal);
		}
	}

	if (--mLifeSpan)
		return true;
	else
		return false;
}


// particle render function
void Particle::render(BITMAP *bmp)
{
	switch (mType) {

		case TYPE_THRUST:
			line (bmp, mOldPos.x + scroll.x - scroll_vel.x, mOldPos.y + scroll.y - scroll_vel.y, mPos.x + scroll.x, mPos.y + scroll.y, yellow);
			break;

		case TYPE_SPARK:
			line (bmp, mOldPos.x + scroll.x - scroll_vel.x, mOldPos.y + scroll.y - scroll_vel.y, mPos.x + scroll.x, mPos.y + scroll.y, white);
			break;

		case TYPE_FLOWER:
			draw_sprite (bmp, mFlower->bmp(mLifeSpan/10.0), mPos.x + scroll.x - mFlower->mCentre, mPos.y + scroll.y - mFlower->mCentre);
			break;
	}
}



// trigger function
void trigger_particles(TRIGGER_TYPE type, const Vector &pos, const Vector &vel)
{
	trigger_sound(type, pos, vel);

	if (!use_particles) return;

	switch (type) {

		case TRIGGER_PLAYER_WALL:
		case TRIGGER_ARROW_WALL:
			for (int i = 0; i < 1; i++) {
				Vector nvel = ~vel;
				Vector pvel (nvel.y, -nvel.x);
				Vector v = nvel * (rand() & 0xff) / 255.0 + pvel * ((rand() & 0xff) - 0x80) / 127.0;
//				v.x += ((rand() & 0xff) - 0x80) / 127.0;
//				v.y += ((rand() & 0xff) - 0x80) / 127.0;
				particles.add (Particle (pos, v, 10, 0.01, Particle::TYPE_SPARK, true));
			}
			break;

		case TRIGGER_ARROW_ROCKET:
			for (int i = 0; i < 4; i++) {
				Vector v = vel;
				v.x += 1 * ((rand() & 0xff) - 0x80) / 127.0;
				v.y += 1 * ((rand() & 0xff) - 0x80) / 127.0;
				v.y -= 1;
				particles.add (Particle (pos, v, 60+i*3, 0.05, Particle::TYPE_FLOWER, true, flower_blue_orange));
			}
			break;

		case TRIGGER_ROCKET_WALL:
	    case TRIGGER_ROCKET_SILENTDEATH:
			for (int i = 0; i < 4; i++) {
				Vector v = vel;
				v.x += 1 * ((rand() & 0xff) - 0x80) / 127.0;
				v.y += 1 * ((rand() & 0xff) - 0x80) / 127.0;
				v.y -= 1;
				particles.add (Particle (pos, v, 60+i*3, 0.05, Particle::TYPE_FLOWER, true, flower_red_yellow));
			}
			break;

		case TRIGGER_ROCKET_PERSON:
			for (int i = 0; i < 50; i++) {
				Vector v (((rand() & 0xff) - 0x80) / 127.0 * 3, ((rand() & 0xff) - 0x100) / 127.0 * 3);
				Flower *flower = flower_purple_red;
				if ((i & 3) == 1)
					flower = flower_blue_purple;
				else if ((i & 3) == 2)
					flower = flower_magenta_violet;
				particles.add (Particle (pos, v, 60+i, 0.05, Particle::TYPE_FLOWER, true, flower));
			}
			break;

		case TRIGGER_KILLED_PERSON:
			for (int i = 0; i < 100; i++) {
				Vector v (((rand() & 0xff) - 0x80) / 127.0 * 3, ((rand() & 0xff) - 0x100) / 127.0 * 3);
				Flower *flower = flower_purple_red;
				if ((i & 3) == 1)
					flower = flower_blue_purple;
				else if ((i & 3) == 2)
					flower = flower_magenta_violet;
				particles.add (Particle (pos, v, 250+i, 0.05, Particle::TYPE_FLOWER, true, flower));
			}
			break;

		case TRIGGER_PLAYER_ROCKET:
			for (int i = 0; i < 2; i++) {
				Vector v (0,0);
				v.x += 1 * ((rand() & 0xff) - 0x80) / 127.0;
				v.y += 1 * ((rand() & 0xff) - 0x80) / 127.0;
				v.y -= 1;
				particles.add (Particle (pos, v, 60+i*3, 0.05, Particle::TYPE_FLOWER, true, flower_blue_orange));
			}
			for (int i = 0; i < 2; i++) {
				Vector v(0,0);
				v.x += 1 * ((rand() & 0xff) - 0x80) / 127.0;
				v.y += 1 * ((rand() & 0xff) - 0x80) / 127.0;
				v.y -= 1;
				particles.add (Particle (pos, v, 60+i*3, 0.05, Particle::TYPE_FLOWER, true, flower_red_yellow));
			}
			break;

		case TRIGGER_FIRE_ARROW:
		case TRIGGER_FIRE_ROCKET:
			break;
	}
}




//static unsigned char *yellow;

// generate graphics etc
void init_particles (void)
{
//	yellow = make_gradient (255, 255, 0);

	flower_red_yellow = new Flower (6, 4, 4, 5, 255, 0, 0, 255, 255, 0);

	flower_blue_orange = new Flower (6, 4, 4, 5, 0, 128, 255, 255, 128, 0);

	flower_purple_red = new Flower (10, 7, 6, 6, 96, 0, 255, 255, 0, 0);

	flower_blue_purple = new Flower (10, 5, 4, 7, 0, 0, 255, 128, 0, 255);

	flower_magenta_violet = new Flower (12, 4, 4, 10, 255, 0, 255, 128, 0, 255);


	yellow = makecol (192, 192, 0);
	white = makecol (255, 255, 255);
}



