// the sound system (a totally generic modular synth, hohum)

#include "sound.h"
#include "level.h"
#include "heart.h"



// generates a random number ranging -1 to 1
#define RAND   (((float)(rand() & 255) / 128.0) - 1.0)



// what frequency are we working at?
#define FREQ	22050



// list of samples
static SAMPLE *engine;
static SAMPLE *fire_arrow;
static SAMPLE *fire_rocket;
static SAMPLE *player_bounce;
static SAMPLE *arrow_bounce;
static SAMPLE *little_explode;
static SAMPLE *medium_explode;
static SAMPLE *big_explode;
static SAMPLE *menu;
static SAMPLE *whoosh;



// abstract base class for synth components that output generated values
class syn_Generator
{
protected:
	syn_Generator() { mLastTick = -1; mLastValue = 0; }

	virtual float read() = 0;

	int mLastTick;

	float mLastValue;

	static int mTickCount;

	static float mSamplePos;
	static float mSampleLen;

	friend class syn_Value;

	friend SAMPLE *synthesize(syn_Generator &synth, float len);
};



// these are shared by all generators
int syn_Generator::mTickCount = 0;

float syn_Generator::mSamplePos = 0;
float syn_Generator::mSampleLen = 0;



// generic data source: holds either a constant value or pointer to a generator
class syn_Value
{
public:
	
	// constructor for constant values
	syn_Value(float value)
		: mValue(value),
		  mGenerator(NULL)
		{ }


	// stupid compiler!
	syn_Value(int value)
		: mValue(value),
		  mGenerator(NULL)
		{ }


	// still a stupid compiler!
	syn_Value(double value)
		: mValue(value),
		  mGenerator(NULL)
		{ }


	// constructor for generated values
	syn_Value(syn_Generator *generator)
		: mValue(0),
		  mGenerator(generator)
		{ }


	// reads the current value of the source
	operator float()
	{
		if (mGenerator) {
			if (mGenerator->mLastTick != mGenerator->mTickCount) {
				mGenerator->mLastValue = mGenerator->read();
				mGenerator->mLastTick = mGenerator->mTickCount;
			}
			
			mValue = mGenerator->mLastValue;
		}

		return mValue;
	}
	
	
protected:
	// data
	float mValue;

	syn_Generator *mGenerator;
};



// square wave oscillator class
class syn_Square : public syn_Generator
{
public:
	
	// constructs the oscillator
	syn_Square(syn_Value freq, syn_Value base = 0, syn_Value depth = 1, syn_Value pulse_width = 0.5)
		: mFreq(freq),
		  mBase(base),
		  mDepth(depth),
		  mPulseWidth(pulse_width)
	{
		mPos = 0;
	}


	// update function
	float read()
	{
		mPos += mFreq / FREQ;

		float v = mPos - (int)mPos;

		if (v > mPulseWidth)
			return mBase + mDepth;
		else
			return mBase -mDepth;
	}
	
	
protected:
	// input controls
	syn_Value mFreq;
	syn_Value mBase;
	syn_Value mDepth;
	syn_Value mPulseWidth;


	// oscillator state
	float mPos;
};



// saw wave oscillator class
class syn_Saw : public syn_Generator
{
public:
	
	// constructs the oscillator
	syn_Saw(syn_Value freq, syn_Value base = 0, syn_Value depth = 1, syn_Value dir = 1)
		: mFreq(freq),
		  mBase(base),
		  mDepth(depth),
		  mDir(dir)
	{
		mPos = 0;
	}


	// update function
	float read()
	{
		mPos += mFreq / FREQ;

		float v = mPos - (int)mPos;
		
		if (mDir < 0)
			v = 1-v;
		
		return mBase + (v*2 - 1) * mDepth;
	}
	
	
protected:
	// input controls
	syn_Value mFreq;
	syn_Value mBase;
	syn_Value mDepth;
	syn_Value mDir;


	// oscillator state
	float mPos;
};



// sine wave oscillator class
class syn_Sine : public syn_Generator
{
public:
	
	// constructs the oscillator
	syn_Sine(syn_Value freq, syn_Value base = 0, syn_Value depth = 1)
		: mFreq(freq),
		  mBase(base),
		  mDepth(depth)
	{
		mPos = 0;
	}


	// update function
	float read()
	{
		mPos += 2 * M_PI * mFreq / FREQ;

		return mBase + sin(mPos) * mDepth;
	}
	
	
protected:
	// input controls
	syn_Value mFreq;
	syn_Value mBase;
	syn_Value mDepth;


	// oscillator state
	float mPos;
};



// noise oscillator class
class syn_Noise : public syn_Generator
{
public:
	
	// constructs the oscillator
	syn_Noise(syn_Value pink = 1, syn_Value base = 0, syn_Value depth = 1)
		: mPink(pink),
		  mBase(base),
		  mDepth(depth)
		{ }


	// update function
	float read()
	{
		float r = RAND;

		if (r < 0)
			r = -pow(-r, mPink);
		else
			r = pow(r, mPink);

		return mBase + r * mDepth;
	}
	
	
protected:
	// input controls
	syn_Value mPink;
	syn_Value mBase;
	syn_Value mDepth;
};



// smooths out high frequency changes
class syn_Damper : public syn_Generator
{
public:
	
	// constructs the object
	syn_Damper(syn_Value input, syn_Value damping = 0, bool reset = true)
		: mInput(input),
		  mDamping(damping)
	{
		mVal = 0;
		mFirst = reset;
	}


	// update function
	float read()
	{
		float v = mInput;
		float d = mDamping;
		
		if (mFirst) {
			mVal = v;
			mFirst = false;
		}
		else
			mVal = mVal*d + v*(1-d);
		
		return mVal;
	}
	
	
protected:
	// input controls
	syn_Value mInput;
	syn_Value mDamping;


	// damping state
	float mVal;
	
	bool mFirst;
};



// only allows signal changes at periodic intervals
class syn_Quantise : public syn_Generator
{
public:
	
	// constructs the object
	syn_Quantise(syn_Value input, syn_Value step = 0)
		: mInput(input),
		  mStep(step)
	{
		mVal = 0;
		mTime = 0;
	}


	// update function
	float read()
	{
		float v = mInput;

		if (mTime <= 0) {
			mVal = v;
			mTime += mStep;
		}
		
		mTime--;
		
		return mVal;
	}
	
	
protected:
	// input controls
	syn_Value mInput;
	syn_Value mStep;


	// damping state
	float mVal;

	float mTime;
};



// multiplies two signals together
class syn_Mul : public syn_Generator
{
public:
	
	// constructs the object
	syn_Mul(syn_Value input1, syn_Value input2)
		: mInput1(input1),
		  mInput2(input2)
		{ }


	// update function
	float read()
	{
		return mInput1 * mInput2;
	}
	
	
protected:
	// input controls
	syn_Value mInput1;
	syn_Value mInput2;
};



// adds multiple signals together
class syn_Mixer : public syn_Generator
{
public:
	
	// constructs an empty object
	syn_Mixer()
	{
	}


	// constructs an object mixing a pair of signals
	syn_Mixer(syn_Value input1, syn_Value input2, syn_Value level1 = 1, syn_Value level2 = 1)
	{
		add(input1, level1);
		add(input2, level2);
	}


	// adds a new input
	void add(syn_Value input, syn_Value level = 1)
	{
		mInputs.push_back(Input(input, level));
	}


	// update function
	float read()
	{
		float v = 0;
		
		for (list<Input>::iterator it = mInputs.begin(); it != mInputs.end(); it++)
			v += it->mInput * it->mLevel;
		
		return v;
	}
	
	
protected:
	// list of input sources
	struct Input
	{
		Input(syn_Value input, syn_Value level)
			: mInput(input),
			  mLevel(level)
			{ }

		syn_Value mInput;
		syn_Value mLevel;
	};

	list<Input> mInputs;
};



// envelope generator
class syn_Envelope : public syn_Generator
{
public:
	
	// constructs the object
	syn_Envelope(syn_Value delay, syn_Value attack, syn_Value decay, syn_Value hold, syn_Value release, syn_Value max = 1, syn_Value mid = 1, syn_Value min = 0)
		: mDelay(delay),
		  mAttack(attack),
		  mDecay(decay),
		  mHold(hold),
		  mRelease(release),
		  mMax(max),
		  mMid(mid),
		  mMin(min)
		{ }


	// alternate constructor for simple attack/release envelopes
	syn_Envelope(syn_Value attack, syn_Value release, syn_Value max = 1, syn_Value min = 0)
		: mDelay(0),
		  mAttack(attack),
		  mDecay(0),
		  mHold(0),
		  mRelease(release),
		  mMax(max),
		  mMid(max),
		  mMin(min)
		{ }


	// update function
	float read()
	{
		float delay = mDelay;
		float attack = mAttack;
		float decay = mDecay;
		float hold = mHold;
		float release = mRelease;
		float max = mMax;
		float mid = mMid;
		float min = mMin;
		
		if (hold < 0) {
			hold = mSampleLen - delay - attack - decay - release;
			
			if (hold < 0)
				hold = 0;
		}
		
		// inital delay period
		if (mSamplePos < delay)
			return min;

		// attack envelope
		if (mSamplePos < delay+attack) {
			float v = (mSamplePos-delay) / attack;
			return min*(1-v) + max*v;
		}

		// decay envelope
		if (mSamplePos < delay+attack+decay) {
			float v = (mSamplePos-delay-attack) / decay;
			return max*(1-v) + mid*v;
		}

		// hold level
		if (mSamplePos < delay+attack+decay+hold)
			return mid;

		// release envelope
		if (mSamplePos < delay+attack+decay+hold+release) {
			float v = (mSamplePos-delay-attack-decay-hold) / release;
			return mid*(1-v) + min*v;
		}

		// finished
		return min;
	}
	
	
protected:
	// input controls
	syn_Value mDelay;
	syn_Value mAttack;
	syn_Value mDecay;
	syn_Value mHold;
	syn_Value mRelease;
	syn_Value mMax;
	syn_Value mMid;
	syn_Value mMin;
};



// creates a sample
SAMPLE *synthesize(syn_Generator &synth, float len)
{
	int samples = (int)(len * FREQ);

	SAMPLE *sample = create_sample(16, false, FREQ, samples);

	unsigned short *data = (unsigned short *)sample->data;

	syn_Generator::mTickCount = 0;
	syn_Generator::mSampleLen = len;

	for (int i=0; i<samples; i++) {
		syn_Generator::mTickCount++;
		syn_Generator::mSamplePos = (float)i / (float)FREQ;

		float val = (synth.read() + 1) * 32768;

		data[i] = MID(0, (int)val, 65535);
	}

   	return sample;
};




// creates a little silent sample
static SAMPLE *synthesize_silence()
{
	syn_Sine osc(0, 0, 0);

	return synthesize(osc, 0.1);
}



// creates the engine sound
static SAMPLE *synthesize_engine()
{
	syn_Saw osc(75);

	syn_Noise noise(4);
	syn_Damper damped_noise(&noise, 0.9);

	syn_Mixer mixer(&osc, &damped_noise, 0.09, 0.666);
	
	return synthesize(mixer, 1);
}



// creates the fire-an-arrow sound
static SAMPLE *synthesize_fire_arrow()
{
	// sine wave with random pitch variation, plus gradual pitch change
	syn_Noise osc1_noise(1, 900, 50);
	syn_Quantise osc1_quantise(&osc1_noise, 100);
	syn_Damper osc1_damper(&osc1_quantise, 0.9, true);
	syn_Envelope osc1_pitch(0, 0.3, 500, &osc1_quantise);
	syn_Sine osc1(&osc1_pitch);

	// another slightly different sine wave
	syn_Noise osc2_noise(1, 900, 50);
	syn_Quantise osc2_quantise(&osc2_noise, 100);
	syn_Damper osc2_damper(&osc2_quantise, 0.9, true);
	syn_Envelope osc2_pitch(0, 0.3, 1200, &osc2_quantise);
	syn_Sine osc2(&osc2_pitch);

	syn_Envelope osc_env(0.05, 0.25, 0.3);

	// modulate the two sines together
	syn_Mul ampmod(&osc1, &osc2);
	syn_Envelope amp_env(0, 0.3, 0.15);
	
	// mix it all together
	syn_Mixer mixer;
	
	mixer.add(&osc1, &osc_env);
	mixer.add(&osc2, &osc_env);
	mixer.add(&ampmod, &amp_env);
	
	return synthesize(mixer, 0.3);
}



// creates the fire-a-rocket sound
static SAMPLE *synthesize_fire_rocket()
{
	// siren sound goes up with tremolo, pulse width etc controlled by LFO
	syn_Envelope siren_pitch(0, 2, 110, 666);

	syn_Envelope siren_lfo_rate(0, 2, 0, 6);
	syn_Mul siren_lfo_rate_squared(&siren_lfo_rate, &siren_lfo_rate);
	syn_Saw siren_lfo_src(&siren_lfo_rate_squared, 0.5, 0.5);
	syn_Damper siren_lfo(&siren_lfo_src, 0.99);

	syn_Mixer siren_amp_mod(&siren_lfo, 0);
	syn_Mixer siren_pulse_width(&siren_lfo, 0);

	syn_Square siren_osc(&siren_pitch, 0, &siren_amp_mod, &siren_pulse_width);

	syn_Envelope siren_vol(0, 0.1, 0, 0, 1.9, 0.1, 0.1, 0);

	// a trio of descending, tremolo base notes, flanged to fuck
	syn_Envelope base_pitch1(0, 2, 110, 80);
	syn_Envelope base_pitch2(0, 2, 115, 75);
	syn_Envelope base_pitch3(0, 2, 105, 85);

	syn_Envelope base_pulse_width(0, 2, 0, 1);

	syn_Sine base_amp_mod1(5, 1, 0.25);
	syn_Sine base_amp_mod2(6, 1, 0.25);
	syn_Sine base_amp_mod3(7, 1, 0.25);

	syn_Square base_osc1(&base_pitch1, 0, &base_amp_mod1, &base_pulse_width);
	syn_Square base_osc2(&base_pitch2, 0, &base_amp_mod2, &base_pulse_width);
	syn_Square base_osc3(&base_pitch3, 0, &base_amp_mod3, &base_pulse_width);

	syn_Envelope base_damping(0, 2, 0.9, 1);

	syn_Damper base_filter1(&base_osc1, &base_damping, false);
	syn_Damper base_filter2(&base_osc2, &base_damping, false);
	syn_Damper base_filter3(&base_osc3, &base_damping, false);

	syn_Envelope base_vol(0, 0.5, 0, 0, 1.5, 0.1, 0.1, 0);

	// mix it all together
	syn_Mixer mixer;
	
	mixer.add(&siren_osc, &siren_vol);

	mixer.add(&base_filter1, &base_vol);
	mixer.add(&base_filter2, &base_vol);
	mixer.add(&base_filter3, &base_vol);
	
	return synthesize(mixer, 2);
}



// creates the player bouncing off walls or rockets sound
static SAMPLE *synthesize_player_bounce()
{
	// quick burst of noise at the start
	syn_Noise noise(2);
	syn_Damper damped_noise(&noise, 0.666);
	syn_Envelope noise_env(0, 0.5, 1);

	// saw wave with random pitch variation, plus gradual pitch change
	syn_Noise osc1_noise(1, 700, 50);
	syn_Quantise osc1_quantise(&osc1_noise, 500);
	syn_Damper osc1_damper(&osc1_quantise, 0.9, true);
	syn_Envelope osc1_pitch(0, 0.5, 500, &osc1_quantise);
	syn_Saw osc1(&osc1_pitch);

	// another slightly different saw wave
	syn_Noise osc2_noise(1, 700, 50);
	syn_Quantise osc2_quantise(&osc2_noise, 500);
	syn_Damper osc2_damper(&osc2_quantise, 0.9, true);
	syn_Envelope osc2_pitch(0, 0.5, 850, &osc2_quantise);
	syn_Saw osc2(&osc2_pitch);

	syn_Envelope osc_env(0.25, 0.25, 0.1);

	// modulate the two saws together
	syn_Mul ampmod(&osc1, &osc2);
	syn_Envelope amp_env(0.1, 0.4, 0.2);

	// a bit of bass would be good...
	syn_Saw bass(80);
	syn_Envelope bass_env(0, 0.5, 0.15);

	// mix it all together
	syn_Mixer mixer;
	
	mixer.add(&damped_noise, &noise_env);
	mixer.add(&osc1, &osc_env);
	mixer.add(&osc2, &osc_env);
	mixer.add(&ampmod, &amp_env);
	mixer.add(&bass, &bass_env);
	
	return synthesize(mixer, 0.5);
}



// creates the arrow bouncing off walls sound
static SAMPLE *synthesize_arrow_bounce()
{
	// quick burst of noise at the start
	syn_Noise noise(2);
	syn_Damper damped_noise(&noise, 0.666);
	syn_Envelope noise_env(0, 0.333, 0.1);

	// sine wave with random pitch variation, plus gradual pitch change
	syn_Noise osc1_noise(1, 500, 50);
	syn_Quantise osc1_quantise(&osc1_noise, 100);
	syn_Damper osc1_damper(&osc1_quantise, 0.9, true);
	syn_Envelope osc1_pitch(0, 0.333, 900, &osc1_quantise);
	syn_Sine osc1(&osc1_pitch);

	// another slightly different sine wave
	syn_Noise osc2_noise(1, 1200, 50);
	syn_Quantise osc2_quantise(&osc2_noise, 100);
	syn_Damper osc2_damper(&osc2_quantise, 0.9, true);
	syn_Envelope osc2_pitch(0, 0.333, 900, &osc2_quantise);
	syn_Sine osc2(&osc2_pitch);

	syn_Envelope osc_env(0.111, 0.222, 0.15);

	// modulate the two sines together
	syn_Mul ampmod(&osc1, &osc2);
	syn_Envelope amp_env(0, 0.333, 0.15);
	
	// a bit of bass would be good...
	syn_Saw bass(80);
	syn_Envelope bass_env(0, 0.333, 0.025);

	// mix it all together
	syn_Mixer mixer;
	
	mixer.add(&damped_noise, &noise_env);
	mixer.add(&osc1, &osc_env);
	mixer.add(&osc2, &osc_env);
	mixer.add(&ampmod, &amp_env);
	mixer.add(&bass, &bass_env);
	
	return synthesize(mixer, 0.333);
}



// creates the small (shoot a rocket) explosion sound
static SAMPLE *synthesize_little_explosion()
{
	// noise at gradually increasing quantisation steps
	syn_Noise noise(2);

	syn_Envelope noise_quantise(0, 1, 0, 35);
	syn_Quantise quantised_noise(&noise, &noise_quantise);

	syn_Envelope noise_damp(0, 1, 0.25, 1);
	syn_Damper damped_noise(&quantised_noise, &noise_damp);

	syn_Envelope noise_env(0, 1, 0.5);

	// a bit of bass would be good...
	syn_Saw bass(80);
	syn_Envelope bass_env(0, 0.5, 0.1);

	// mix it all together
	syn_Mixer mixer;
	
	mixer.add(&damped_noise, &noise_env);
	mixer.add(&bass, &bass_env);
	
	return synthesize(mixer, 1);
}



// creates the medium (rocket hits person) explosion sound
static SAMPLE *synthesize_medium_explosion()
{
	// noise at gradually increasing quantisation steps
	syn_Noise noise;

	syn_Envelope noise_quantise(0, 3, 0, 50);
	syn_Quantise quantised_noise(&noise, &noise_quantise);

	syn_Envelope noise_damp(0, 3, 0.1, 1);
	syn_Damper damped_noise(&quantised_noise, &noise_damp);

	syn_Envelope noise_env(0, 3, 0.5);

	// lots of bass, please!
	syn_Envelope bass_amp_freq(0, 3, 15, 1);
	syn_Sine bass_amp(&bass_amp_freq, 0.5, 0.5);

	syn_Envelope bass_pitch(0, 3, 80, 50);

	syn_Sine bass(&bass_pitch, 0, &bass_amp);

	syn_Envelope bass_env(0, 3, 0.666);

	// mix it all together
	syn_Mixer mixer;
	
	mixer.add(&damped_noise, &noise_env);
	mixer.add(&bass, &bass_env);
	
	return synthesize(mixer, 3);
}



// creates the massive (person dies) explosion sound
static SAMPLE *synthesize_big_explosion()
{
	// noise at gradually increasing quantisation steps
	syn_Noise noise;

	syn_Envelope noise_quantise(0, 7, 0, 50);
	syn_Quantise quantised_noise(&noise, &noise_quantise);

	syn_Envelope noise_damp(0, 7, 0.1, 1);
	syn_Damper damped_noise(&quantised_noise, &noise_damp);

	syn_Envelope noise_env(0, 7, 0.666);

	// lots of bass, please!
	syn_Envelope bass_amp_freq1(0, 6, 15, 1);
	syn_Envelope bass_amp_freq2(0, 6, 16, 1);

	syn_Sine bass_amp1(&bass_amp_freq1, 0.5, 0.5);
	syn_Sine bass_amp2(&bass_amp_freq2, 0.5, 0.5);

	syn_Envelope bass_pitch1(0, 6, 85, 20);
	syn_Envelope bass_pitch2(0, 6, 80, 25);

	syn_Envelope bass_pulse_width(0, 6, 0, 1);

	syn_Square bass1(&bass_pitch1, 0, &bass_amp1, &bass_pulse_width);
	syn_Square bass2(&bass_pitch2, 0, &bass_amp2, &bass_pulse_width);

	syn_Envelope bass_env(0, 6, 0.5);

	// thereminy sine wave type thing
	syn_Noise osc1_noise(1, 0, 100);
	syn_Quantise osc1_quantise(&osc1_noise, 500);
	syn_Damper osc1_damper(&osc1_quantise, 0.95, true);
	syn_Envelope osc1_pitch_env(0, 0.5, 6.5, 0, 0, 1500, 1000, 80);
	syn_Mixer osc1_pitch(&osc1_damper, &osc1_pitch_env);
	syn_Sine osc1(&osc1_pitch);

	// and another...
	syn_Noise osc2_noise(1, 0, 100);
	syn_Quantise osc2_quantise(&osc2_noise, 500);
	syn_Damper osc2_damper(&osc2_quantise, 0.95, true);
	syn_Envelope osc2_pitch_env(0, 1, 6, 0, 0, 1400, 2000, 80);
	syn_Mixer osc2_pitch(&osc2_damper, &osc2_pitch_env);
	syn_Sine osc2(&osc2_pitch);

	syn_Envelope osc_env(0, 1, 0, 0, 6, 0.05, 0.05);

	// mix it all together
	syn_Mixer mixer;
	
	mixer.add(&damped_noise, &noise_env);
	mixer.add(&bass1, &bass_env);
	mixer.add(&bass2, &bass_env);
	mixer.add(&osc1, &osc_env);
	mixer.add(&osc2, &osc_env);
	
	return synthesize(mixer, 7);
}



// creates the menu click sound
static SAMPLE *synthesize_menu()
{
	syn_Sine osc1(666);
	syn_Sine osc2(999);

	syn_Envelope osc_env(0.025, 0.175, 0.1);

	syn_Noise noise;
	syn_Damper damped_noise(&noise, 0.75);
	syn_Envelope noise_env(0, 0.1, 0.05);

	syn_Mixer mixer;
	
	mixer.add(&osc1, &osc_env);
	mixer.add(&osc2, &osc_env);
	mixer.add(&damped_noise, &noise_env);
	
	return synthesize(mixer, 0.2);
}



// creates the whoosh sound
static SAMPLE *synthesize_whoosh()
{
    syn_Noise noise(0.1);

    syn_Sine osc1(100);
    syn_Sine osc2(50, 0.3, 0.3);

    syn_Mixer ampmod;

    ampmod.add(&osc1);
    ampmod.add(&osc2);

    syn_Envelope env(0.5, 0.1, 0.3);

    syn_Mixer mixer;

    mixer.add(&noise, &ampmod);

    syn_Mixer mixer2;

    mixer2.add(&mixer, &env);

    syn_Damper damper(&mixer2, 0.85);

    return synthesize(damper, 3);
}



// once only init at startup
void init_sound()
{
	install_sound(DIGI_AUTODETECT, MIDI_AUTODETECT, NULL);

	engine = synthesize_engine();
	fire_arrow = synthesize_fire_arrow();
	fire_rocket = synthesize_fire_rocket();
	player_bounce = synthesize_player_bounce();
	arrow_bounce = synthesize_arrow_bounce();
	little_explode = synthesize_little_explosion();
	medium_explode = synthesize_medium_explosion();
	big_explode = synthesize_big_explosion();
	menu = synthesize_menu();
    whoosh = synthesize_whoosh();
}



// worker function for playing sounds
static void trigger(SAMPLE *sample, Vector pos, float v = 1, float dscale = 1)
{
	pos += scroll;
	
	float d = magnitude(pos - Vector(SCREEN_W/2, SCREEN_H/2)) * dscale;
	
	d /= (SCREEN_W/4);
	
	if (d > 1)
		d = 1.0 / d;
	else
		d = 1;
	
	int vol = MID(0, (int)(v*d*255), 255);

	int pan = MID(0, (int)(pos.x*255/SCREEN_W), 255);

	int height = MID(-256, (int)(pos.y*256/SCREEN_H-128), 256);
	
	int pitch = 1000 + (int)(RAND*50) - height;
	
	play_sample(sample, vol, pan, pitch, false);
}



// trigger function
void trigger_sound(TRIGGER_TYPE type, const Vector &pos, const Vector &vel)
{
	switch (type) {
	
		case TRIGGER_PLAYER_WALL:
		case TRIGGER_PLAYER_ROCKET:
			{
				float d = heart->mVel.magnitude();

				if (type == TRIGGER_PLAYER_WALL) {
					d *= ABS(vel | ~heart->mVel);
					d -= 0.333;
				}
				else
					d += 1;

				if (d > 0)
					trigger(player_bounce, pos, d*0.25);
			}
			break;

		case TRIGGER_FIRE_ARROW:
			trigger(fire_arrow, pos);
			break;

		case TRIGGER_ARROW_WALL:
			trigger(arrow_bounce, pos);
			break;

		case TRIGGER_ARROW_ROCKET:
		case TRIGGER_ROCKET_WALL:
			trigger(little_explode, pos);
			break;

		case TRIGGER_FIRE_ROCKET:
			trigger(fire_rocket, pos);
			break;

		case TRIGGER_ROCKET_PERSON:
			trigger(medium_explode, pos, 0.8, 0.5);
			break;

		case TRIGGER_KILLED_PERSON:
			trigger(big_explode, pos, 0.8, 0.25);
			break;
	}
}



// engine status
static int engine_vol = 0;

static int fwds = true;



// updates the engine sound
void update_engine(bool forward, bool reverse)
{
	int prev_vol = engine_vol;

	if (forward || reverse)
		engine_vol = MIN(engine_vol+50, 255);
	else
		engine_vol = MAX(engine_vol-10, 0);

	int pitch = 850 + (int)(heart->mVel.magnitude()*100);

	if (forward)
		fwds = true;
	else if (reverse)
		fwds = false;

	if (!fwds)
		pitch += 666;

	if (engine_vol) {
		if (!prev_vol)
			play_sample(engine, engine_vol, 128, pitch, true);
		else
			adjust_sample(engine, engine_vol, 128, pitch, true);
	}
	else if (prev_vol)
		stop_sample(engine);
}



// stops the engine sound
void stop_engine()
{
	if (engine_vol) {
		stop_sample(engine);
		engine_vol = 0;
	}
}



// plays the menu click sound
void menu_click(bool big)
{
	play_sample(menu, (big) ? 255 : 192, 128, (big) ? 900 : 1200, false);
}



// plays the whoosh when someone comes on/off the screen
void player_whoosh()
{
    play_sample(whoosh, 64, 128, 1500, false);
}


